//////////////////////////////////////////////////////////
// Parralel LU factorization
// Kristian Kvilekval
// cs240b - spring98

#include <stdlib.h>
#include <iostream.h>
#include <float.h>
#include <getopt.h>

#include "mpi.h"

//#ifdef __mips
//#  include <cblas.h>
//#endif
#  define BLAS(X) X##_

#define DEBUX(X)

extern "C" 
{
  int dtrsm_(...);
  int dtrsv_(...);
  int dgemm_(...);
  int dgemv_(...);
}

template<class EL>
class array2D
{
public: 
  typedef EL value_type;

public:
  array2D () : rows_(0), cols_(0), els_(0) {}
  array2D (int rows, int cols) { allocate(rows, cols); }
  ~array2D () { delete[] els_; }

  ///////////////////////////////////
  // info
  int rows() const { return rows_; }
  int cols() const { return cols_; }
  int rank() const { return  (rows_ == cols_)?rows_:-1;}
  long elements () { return (long)rows_*cols_; }

  EL& index (int row, int col) { return els_[row + col*rows_]; }
  const EL& index (int row, int col) const { return els_[row + col*rows_]; }

  void allocate (int rows, int cols) 
    { 
      rows_ = rows; cols_ = cols;  els_ = new EL[cols*rows]; 
      DEBUX (cerr << "created array : " << cols*rows << endl);
    }

  EL* get_data() const { return els_; }
  long get_data_size () const { return elements() * sizeof (EL); }

  void operator = (const array2D<EL>& m) { assign (m); }
  void operator = (EL m) { assign (m); }

  friend ostream& operator << (ostream&o, array2D<EL>&m)
  {
      for (int r=0; r < m.rows_; r++)
      {
	for (int c=0; c < m.cols_; c++)
	  o << ' ' << m.index(r,c) ;
        o << endl;
      }
      return o;
  }
  
private:
  void assign (EL m) 
    {
      for (int c=0; c < cols_; c++)
	for (int r=0; r < rows_; r++)
	  index(r,c) = m;
    }
  void assign (const array2D<EL>& m)
    {
      for (int i=0; i < elements(); i++)
	els_[i] = m.els_[i];
    }

  int rows_;
  int cols_;
  EL* els_;
};


class lusolver 
{
public:
  typedef array2D<double> block;

  struct swap_pair
  {
    int swap_block;
    int swap_row;
  };

public:
  lusolver(MPI::Intracomm& com, int matrix_sz , int block_size = -1);
  ~lusolver();

  void init();
  void solve ();
  bool check();

  friend ostream& operator <<  (ostream&o, const lusolver &lus) 
    {
      for (int r=0; r < lus.matrix_size; r++)
      {
	if (r) o << ',';
	o << lus.results[r];
      }
      return o;
    }

   void set_pivot (bool x) { do_pivoting = x; }
   void set_debug (bool x) { do_debug = x; }

private:  
  void find_pivot(int cb, int col, swap_pair&pivot);
  void swap_local_row(int cb, int rb, int row, swap_pair& pivot);
  void reduce(int cb, int col);
  void lufactor_block (int cb, swap_pair swaps[]);
  void apply_swaps (int cb, swap_pair swaps[]);
  void llsolve_eliminate (int cb, block & ll);
  void lufactor ();
  void solve_upper ();

  array2D<block> m;

  MPI::Intracomm& comm;
  int block_size;
  int block_count;
  int matrix_size;
  int proc_rank;
  int comm_size;

  double *results;

  bool do_pivoting;
  bool do_debug;
};



lusolver::lusolver(MPI::Intracomm& com, int matrix_sz , int block_sz)
  : comm(com),
    matrix_size(matrix_sz)
{
  comm_size = comm.Get_size();
  proc_rank = comm.Get_rank();

  DEBUX(cerr << proc_rank << ": size :" << comm_size << endl);

  if (block_sz < 0)
    {
      block_size = matrix_size / comm_size;
    }
  else 
    {
      block_size = block_sz; 
    }
  
  results = 0;
  do_pivoting = true;
  do_debug = false;

  block_count = matrix_size / block_size;

  m.allocate(block_count, block_count);
}

lusolver::~lusolver() 
{ 
  delete[] results; 
}


/////////////////////////////////////
// initialize each column that this
// node owns.
// Initial matrix layout is all 1's
// with a N running down the reverse
// diaganol
void lusolver::init()
{
  DEBUX(cerr << proc_rank << " : init" << endl);

  for (int cb=0; cb<block_count; cb++)
    if (cb % comm_size == proc_rank)
      for (int rb=0; rb<block_count; rb++)
	{
	  m.index(rb, cb).allocate (block_size, block_size);

	  // Generally initialize every matrix to 1.
	  m.index(rb, cb) = 1;

	  // Place on N on the reverse diaganol
	  if (cb == block_count - rb -1 )
	    for (int r=0; r < block_size; r++)
	      m.index(rb, cb).index (r, block_size - r - 1) = matrix_size;
	}

  results = new double[matrix_size];
  if (proc_rank == 0) 
    {
      // maintain the solution values;

      for (int r = 0; r < matrix_size; r++)
	results[r] = (double)2*matrix_size - 1;
    }
}

bool lusolver::check()
{
  
  for (int r = 0; r < matrix_size; r++)
    {
      if (results[r] - 1.0 > 1000*DBL_EPSILON 
	  || 1.0 - results[r] > 1000 *DBL_EPSILON) 
	{
	  cerr << "bad at " << r << " value " << results[r] << endl;
	  return false;
	}
    }
  return true;
}

//////////////////////////////////////////
// Find the the pivot value for a given 
// column
void lusolver::find_pivot(int cb, int col, swap_pair&pivot)
{
  int rb = cb;
  double pivot_val = -DBL_MAX;

  // Scan each rowblock
  for (rb=cb; rb < block_count; rb++)
    {
      block& b = m.index(rb,cb);

      //scan each block;
      // The first block starts down a little, the rest from row 0
      for (int row = (cb==rb)?col:0; row < block_size; row++)
	if (b.index (row,col) > pivot_val)
	  {
	    pivot.swap_block = rb;
	    pivot.swap_row = row;
	    pivot_val = b.index (row,col);
	  }
    }
}

////////////////////////////////////
// Interchange a row that is possibly
// in another block.
void lusolver::swap_local_row(int cb, int rb, int row, swap_pair& pivot)
{
  int c;
  double tmp;

  // Don't swap if nothing moved.
  if (rb == pivot.swap_block && row == pivot.swap_row)
    return;

//  DEBUX (cerr << proc_rank << " : swapping ("  << rb << ',' << row << ')'
//				 << '(' << pivot.swap_block << ',' 
//			<< pivot.swap_row << ')' <<  cb << endl);
  for (c=0; c < block_size; c++)
    {
      tmp= m.index(rb,cb).index(row, c);

      m.index(rb, cb).index(row, c) =
	m.index(pivot.swap_block, cb).index(pivot.swap_row, c);

      m.index(pivot.swap_block, cb).index(pivot.swap_row, c) = tmp;
    }
}

//////////////////////////////////////////
// Reduce this block (expect that 
// pivoting has been done).
void lusolver::reduce(int cb, int col)
{
  int orig_rb = cb;
  int orig_row = col;
  //int orig_col = col;

  //??? Can this be replaced by a BLAS routine????

  // Reduce the entire column by scanning the row blocks.
  for (int rb=cb; rb < block_count; rb++)
    {
      block& block = m.index (rb,cb);
      // scan the rows in this block.
      for (int row = (rb==cb)?orig_row+1:0; row  < block_size; row++)
	{
	  // Calculate factor. a_ik /= a_kk
	  block.index (row,col) /=  
	    m.index(orig_rb, cb).index (orig_row, col);
	  
	  // Apply factor to rest of row in this block.
	  for (int c=col+1; c < block_size; c++)
	    {
	      // a_ij = a_ij - a_ik * a_kj
	      block.index(row, c) -= 
		block.index (row,col) * m.index(orig_rb,cb).index (orig_row,c);
	    }
	}
    }
}

/////////////////////////////////////////
// Factor the block by going through
// each column: finding a pivot, swapping
// reducing the column
void lusolver::lufactor_block (int cb, swap_pair swaps[])
{
  for (int col = 0; col < block_size; col++)
    {
      if (do_pivoting)
      {
        find_pivot (cb, col, swaps[col]);
        swap_local_row (cb, cb, col, swaps[col]);
      }

      reduce (cb, col);
    }
}

////////////////////////////////////
// Apply swaps to other columns
void lusolver::apply_swaps (int cb, swap_pair swaps[])
{
  int rb=cb;
  // send out the swapped rows
  DEBUX(cerr << proc_rank << " : > apply swaps" << endl);
  comm.Bcast (swaps, block_size, MPI::TWOINT, cb % comm_size);

  for (int xcb=0; xcb < block_count; xcb++)
    if (xcb  != cb)
      {
	// If I wasn't the sender then I need to do the swaps here.
	// There can be at most block_size swaps
	if (xcb % comm_size == proc_rank)
	  for (int row=0; row < block_size; row++)
	    swap_local_row (xcb, rb, row, swaps[row]);
      }

  // pivot answer array
  if (proc_rank == 0)
    {
      for (int row=0; row < block_size; row++)
	{
	  int indx = swaps[row].swap_block*block_size + swaps[row].swap_row;
	  double tmp;
	  tmp = results[indx];
	  results[indx] = results[rb*block_size+row];
	  results[rb*block_size+row] = tmp;
	}
    }
  DEBUX(cerr << proc_rank << " :  < apply swaps" << endl);
}

//////////////////////////////////////
// Use the reduced lower triangular
void lusolver::llsolve_eliminate (int cb, block & ll)
{
  int rb = cb;
  int orig_rb = rb;
  double one = 1.0;
  double minus_one = -1.0;
  int int_one = 1;

  // Send the lower triangular matrix to the other columns
  if (cb % comm_size ==  proc_rank)
    ll = m.index (rb, cb);

  // I really want these only to the right and maybe to proc 0 
  // but the other procs are doing nothing anyway.
  comm.Bcast (ll.get_data(), ll.elements(), MPI::DOUBLE, cb % comm_size);

  // Only those to the right of the current block need to factor
  for (int xcb = cb+1; xcb < block_count; xcb++)
    if (xcb % comm_size == proc_rank)
      {
	// B(rb,rb) = LL' * B(rb,rb);
	DEBUX (cerr << proc_rank << " : llsolve elim" << endl);
	BLAS(dtrsm)("Left", "Lower", "NoTranspose", "Unit",
		    &block_size, &block_size, &one,
		    ll.get_data(), &block_size,
		    m.index(rb, xcb).get_data(), &block_size);
      }

  if (proc_rank == 0)
    {
      // If proc 0 then solve in b.
      // b = LL' * b;
      BLAS(dtrsv)("Lower","NoTran","Unit",
		  &block_size,
		  ll.get_data(),&block_size,
		  &results[cb*block_size],&int_one);
    }

  // send factors and eliminate 
  for (rb++; rb < block_count; rb++)
    {
      if (cb % comm_size == proc_rank)
	ll = m.index(rb, cb);

      // Recieve the factor matrix from the right.
      comm.Bcast (ll.get_data(), ll.elements(), MPI::DOUBLE, cb % comm_size);

      // Only eliminate those blocks beyond the current working block.
      for (int xcb=cb+1; xcb < block_count; xcb++)
	if (xcb % comm_size == proc_rank)
	  {
	    // B(rb,cb) = B(rb, cb) - 1FF * B(origrb,cb) 
	    BLAS(dgemm)("NoTranspose","NoTranspose",
		      &block_size, &block_size, &block_size, 
			&minus_one, ll.get_data(), &block_size,
			m.index(orig_rb, xcb).get_data(), &block_size, 
			&one, m.index(rb, xcb).get_data(), &block_size);
	  }
      
      if (proc_rank == 0)
	{
	  // Update result vector
	  // b[rb] = b[rb] - FF*b[orig_rb];
	  BLAS(dgemv)("NoTrans", 
		      &block_size, &block_size,
		      &minus_one, ll.get_data(), &block_size,
		      &results[orig_rb*block_size], &int_one, 
		      &one, &results[rb*block_size], &int_one);
	}
    }
}

void lusolver::lufactor ()
{
  swap_pair *swaps = new swap_pair[block_size];
  block ll (block_size, block_size); // tmp

  for (int cb=0; cb < block_count; cb++)
    {
      if (cb % comm_size == proc_rank)
	{
	  // My turn to factor block.

	  lufactor_block (cb, swaps);
// 	  if (do_debug) 
// 	  {
// 	    cerr << "factored :" << endl << m.index (cb,cb) << endl;
// 	  }
	}

      if (do_pivoting)
        apply_swaps (cb, swaps);

      llsolve_eliminate(cb, ll);
//       if (do_debug) 
// 	  {
// 	    cerr << "solve-elim:" << endl << m.index (cb,cb) << endl;
// 	  }
    }
  delete [] swaps;
}


void lusolver::solve_upper ()
{
  MPI::Status status;
  int int_one = 1;
  double one = 1.0;
  double minus_one = -1.0;

  int tag = 100;

  DEBUX(cerr << proc_rank << ": > solver_upper" << endl);

  if (proc_rank == 0)
    {
      // Send the original result matrix.
      comm.Send (results, matrix_size, MPI::DOUBLE, (block_count-1)%comm_size, tag);
    }
  if ((block_count-1)%comm_size == proc_rank)
    {
      comm.Recv (results, matrix_size, MPI::DOUBLE, 0, tag);
    }

  for (int cb=block_count-1; cb>=0; cb--)
    if (cb % comm_size == proc_rank)
      {
	int rb = cb;
	int next = cb+1;
	int prev = cb-1;

	if (prev < 0) prev = MPI::PROC_NULL;

	if (next == block_count) 
	  next = MPI::PROC_NULL;
	else
	  next %= comm_size;
	if (prev < 0) prev = MPI::PROC_NULL;
	else
	  prev %= comm_size;

	comm.Recv (results, matrix_size, MPI::DOUBLE, next, tag);

	// Solve the triangular system for this block.

	// b[rb] = B(rb,cb)' * b[rb]
	BLAS(dtrsv)("Upper","NoTranspose","NonUnit", 
		    &block_size,
		    m.index (rb,cb).get_data(), &block_size,
		    &results[rb*block_size],&int_one);

	// Substitute found values into other blocks.
	for (int xrb=rb-1; xrb>=0; xrb-- )
	  {
	    // b[rb] = b[rb] - B(rb,cb) * b[xrb]
	    BLAS(dgemv)("NoTranspose", 
			&block_size, &block_size, 
			&minus_one, m.index(xrb,cb).get_data(), &block_size,
			&results[cb*block_size],&int_one,
			&one, &results[xrb*block_size],&int_one);
	  }
  
	comm.Send (results, matrix_size, MPI::DOUBLE, prev, tag);
      }

  DEBUX(cerr << proc_rank << ": < solver_upper" << endl);
}


void lusolver::solve ()
{
  // LUFactor the matrix

  lufactor();

  solve_upper ();
}

void thread_main(int argc, char **argv)
{
  int rank, size;
  double starttime, endtime;
  int block_size = 25;
  int debug = 0;
  int n;

  MPI::Init (argc, argv);

  size = MPI::COMM_WORLD.Get_size();
  rank = MPI::COMM_WORLD.Get_rank();

  if (rank == 0)
    {
      if (argc < 2)
	{
	  cerr << "usage:plu [-b block-size] <matrix_size>" << endl;
	  exit(0);
	}

      int c;
      while ((c = getopt (argc,argv,"db:")) != -1)
	switch (c)
	  {
	  case 'b': block_size = atoi (optarg);
	    break;
	  case 'd': debug++; 
	    break;
	  default:
	    cerr << "usage:plu [-b block-size] <matrix_size>" << endl;
	    exit(0);
	  }

      n = atoi (argv[optind]);
    }


  MPI::COMM_WORLD.Bcast (&n, 1, MPI::INT, 0);

  // Initialize the blocks that are on this node.

  DEBUX(cerr << rank << ": create : " << n << endl);


  lusolver bm(MPI::COMM_WORLD, n, block_size);

  bm.init();

  if (rank == 0)
    {
      cout << "solving " << n << " by "  << n 
	   << " matrix with block_size " << block_size 
	   << " on " <<size << "procs" << endl;
      starttime = MPI::Wtime();
    }

  bm.solve();

  if (rank == 0)
    {
      endtime = MPI::Wtime();

      cout << "solved " << n << " by " << n << " system in " 
	   << endtime - starttime << endl;

      if (bm.check() == false)
	{
	  cout << "check was bad" << endl;
	  cout << "result was [" << bm << "]" << endl;
	}
      
    }

  MPI::Finalize();  
}


main(int argc, char **argv)
{
  MPI::Init (argc, argv, thread_main);
}
